// Copyright 2019-2024 Yuji Tanaka
// HP : http://galaxystar.image.coocan.jp/
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
/*
 * ESP32S3 I2S player by multi task for ILI9163
 *   Support PSRAM
 *   
 *   SD card : "*.wav"
 *       Linear PCM, 44.1kHz, Signed 16bit, Stereo
 *
 * 2024'05/19 Yuji Tanaka 
 */
#include <WiFi.h>
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <ILI9163_buf.h>
#include <driver/i2s.h>

#define TFT_DC  48
#define TFT_CS  47
#define TFT_RST 45
#define SD_CS   10
#define LED_PIN 40
#define AMP_PIN 41

ILI9163_buf tft = ILI9163_buf(TFT_CS, TFT_DC, TFT_RST);

#define MAX_LINE 16
int _width, _height;

#define MAX_FILE 128
File root;
char fname[MAX_FILE][32];
char fnum;

#define PIN_I2S_BCLK 18
#define PIN_I2S_LRC  19
#define PIN_I2S_DOUT 8
#define SAMPLE_RATE  44100
#define BLOCK_SAMPLES 256
int txbuf[BLOCK_SAMPLES*2];

#define VOL_MAX 16
int  num, old_num;
int  disp_num;
int  vol, old_vol;
int  vol256[VOL_MAX+1]={
    0,  1,  2,  3,  4,  6,   8,  11,
   16, 23, 32, 45, 64, 90, 128, 181,
  256
};

int StickXPin = 6;
int StickYPin = 7;
int StickX;
int StickY;

int ButtonAPin = 15;
int ButtonBPin = 16;
int ButtonA;
int ButtonB;
int OldA;
int OldB;

int ftype;
int I2S_stop;

#define BUF_SIZE (BLOCK_SAMPLES*2*2)
#define BUF_NUM  512
const int offset = 0x2C;
char *data[BUF_NUM];
int data_num[BUF_NUM];
int buf_max;
int r_count;
int w_count;
int w_next;
int file_read;
File file;
char old_fname[32];

TaskHandle_t th;

void setup() {
  int i;
  char str[32];
  
  WiFi.disconnect(true);
  Serial.begin(115200);

  // LCD
  tft.begin(40000000);
  tft.setRotation(2);
  _width=tft.width();
  _height=tft.height();

  // SD card
  Serial.print("Initializing SD card...");
  // pinMode(SD_CS, PULLUP);
  // pinMode(MISO, PULLUP);
  // if (!SD.begin(SD_CS)) {
  if (!SD.begin()) {
    Serial.println("failed!");
    show_no_sd();
    while(1);
  }
  root = SD.open("/");
  strcpy(old_fname, "/");
  pinMode(LED_PIN, OUTPUT);

  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE, ILI9163_BLACK);
  tft.setCursor(_width/2-64, _height/2);
  tft.print("Now read SD card");
  tft.display();

  fnum=0;
  printDirectory(root, 0, ".wav");

  Serial.println("OK!");

  // setup I2S 
  i2s_config_t i2s_config = {
    .mode =  (i2s_mode_t)(I2S_MODE_MASTER | I2S_MODE_TX  | I2S_MODE_RX),
    .sample_rate = SAMPLE_RATE,
    .bits_per_sample = (i2s_bits_per_sample_t)32,
    .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
    .communication_format = (i2s_comm_format_t)(I2S_COMM_FORMAT_STAND_I2S),
    .intr_alloc_flags = 0,
    .dma_buf_count = 6,
    .dma_buf_len = BLOCK_SAMPLES*4,
    .use_apll = false,
    .tx_desc_auto_clear = true,
    .fixed_mclk = 0,
  };
  i2s_driver_install( I2S_NUM_0, &i2s_config, 0, NULL);

  i2s_pin_config_t pin_config = {
        .bck_io_num   = PIN_I2S_BCLK,
        .ws_io_num    = PIN_I2S_LRC,
        .data_out_num = PIN_I2S_DOUT,
        .data_in_num  = -1                                                       
    };
  i2s_set_pin( I2S_NUM_0, &pin_config);
  i2s_zero_dma_buffer(I2S_NUM_0);

  pinMode(AMP_PIN, OUTPUT);
  digitalWrite(AMP_PIN, 0);
  pinMode(ButtonAPin, INPUT_PULLUP);
  pinMode(ButtonBPin, INPUT_PULLUP);

  ftype=0;
  I2S_stop=0;
  num=0;
  old_num=-1;
  disp_num=0;
  vol=8;
  file_read=0;

  xTaskCreatePinnedToCore(main_loop, "main_loop", 8192, NULL, 3, &th, 0);

  if (psramFound()){
    Serial.println("PSRAM valid");
    for(i=0;i<BUF_NUM;i++){
      data[i]=(char *)ps_malloc(BUF_SIZE);
      if (data[i]==NULL)
        break;
    }
  }else{
    Serial.println("PSRAM not valid");
    for(i=0;i<192;i++){
      data[i]=(char *)malloc(BUF_SIZE);
      if (data[i]==NULL)
        break;
    }
  }
  buf_max = i;
  sprintf(str, "%d block RAM", buf_max);
  Serial.println(str);
}

void loop() {
  int i;
  
  if (ftype>=1){
    i=num;
    for(;;){
      I2Splay(fname[i]);
      num=i;
      i++;
      if (i>=fnum)
        i=0;
      if (I2S_stop>=4)
        break;
    }
    ftype=0;
    old_num=-1;
  }
}

void main_loop (void *pvParameters)
{
  int push_btn;
  char str[16];
  int sz;

  while(1) {
    digitalWrite(LED_PIN, 0);
    StickX = analogRead(StickXPin)/4;
    StickY = analogRead(StickYPin)/4;
    OldA=ButtonA;
    OldB=ButtonB;
    ButtonA = digitalRead(ButtonAPin);
    ButtonB = digitalRead(ButtonBPin);

    // Stand-by
    if (ftype==0){
      I2S_stop=0;
      if (StickX<256){
        if (vol>0) vol--;
        vTaskDelay(200);
      }
      if (StickX>768){
        if (vol<VOL_MAX) vol++;
        vTaskDelay(200);
      }
    
      if (StickY<256){
        num++;
        vTaskDelay(200);
      }
      if (StickY>768){
        num--;
        vTaskDelay(200);
      }
      if (num<0)
        num=fnum-1;
      if (num>=fnum)
        num=0;

      if (ButtonA==0 && OldA==1){
        if (num==fnum)
          ftype=2;
        else
          ftype=1;
      }

      if (ButtonB==0 && OldB==1){
        ftype=0;
        num=0;
        old_num=-1;
      }

      if (num==0)
        disp_num=0;
      else if (num==fnum && fnum>=MAX_LINE)
        disp_num=num-MAX_LINE+1;
      else if (num-disp_num>=(MAX_LINE-1))
        disp_num=num-MAX_LINE+2;
      else if (num-disp_num<=0)
        disp_num=num-1;

      if (num!=old_num)
        show_wav_sel();

    // Play
    }else if (ftype==1 || ftype==2){
      // I2Splay
      if (StickX<256){
        if (vol>0) vol--;
        vTaskDelay(200);
      }
      if (StickX>768){
        if (vol<VOL_MAX) vol++;
        vTaskDelay(200);
      }

      if (ButtonB==0 && OldB==1){
        I2S_stop=1;
        file_read=0;
      }
    }else{
      ftype=0;
    }

    if (vol!=old_vol){
      tft.setTextSize(1);
      tft.setTextColor(ILI9163_WHITE, ILI9163_BLACK);
      tft.setCursor(0, _height-9);
      sprintf(str, "Volume : %d  ", vol);
      tft.print(str);
      tft.display();
    }
    old_num=num;
    old_vol=vol;

    // File read
    if (file_read==1 && w_next!=r_count){
      do{
        digitalWrite(LED_PIN, 1);
        // sz=file.readBytes(data[w_count], BUF_SIZE);
        sz=file.read((uint8_t *)data[w_count], BUF_SIZE);
        data_num[w_count]=sz;
        digitalWrite(LED_PIN, 0);
        if (sz==0){
          file_read=0;
        }else{
          w_count++;
          if (w_count>=buf_max)
            w_count=0;
          w_next=w_count+1;
          if (w_next>=buf_max)
            w_next=0;
        }
      // Exit every 16 counts for watchdog.
      }while(w_next!=r_count && (w_next % 16)!=0);
    }else{
      vTaskDelay(10);
    }
  }
}

void printDirectory(File dir, int numTabs, char *typestr) {
  while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    // for (uint8_t i = 0; i < numTabs; i++) {
    //   Serial.print('\t');
    // }
    // Serial.print(entry.name());
    if (fnum<MAX_FILE && stristr(entry.name(), typestr)!=NULL){
      strcpy(fname[fnum], old_fname);
      strcat(fname[fnum++], entry.name());
    }else if(fnum>=MAX_FILE){
      entry.close();
      return;
    }
    
    if (entry.isDirectory()) {
      strcpy(old_fname, "/");
      strcat(old_fname, entry.name());
      strcat(old_fname, "/");
      // Serial.println("/");
      printDirectory(entry, numTabs + 1, typestr);
    // } else {
    //   // files have sizes, directories do not
    //   Serial.print("\t\t");
    //   Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
}

uint16_t read16(File &f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File &f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

char *stristr(const char *org, char *str)
{
  char *tmp, *tmp0, *tmp1;
  tmp=(char *)org;
  for(;*tmp!='\0';tmp++){
    tmp0=tmp;
    tmp1=str;
    for(;*tmp1!='\0';tmp1++){
      if (*tmp0==*tmp1){
        tmp0++;
        continue;
      }else if ('A'<=*tmp1 && *tmp1<='Z' && *tmp0==(*tmp1-'A'+'a')){
        tmp0++;
        continue;
      }else if ('a'<=*tmp1 && *tmp1<='z' && *tmp0==(*tmp1-'a'+'A')){
        tmp0++;
        continue;
      }else
        break;
    }
    if (*tmp1=='\0')
      return(tmp);
  }
  return(NULL);
}

void show_no_sd()
{
  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_RED, ILI9163_BLACK);
  tft.setCursor(16, 0);
  tft.print("Error");

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE, ILI9163_BLACK);
  tft.setCursor(0, 32);
  tft.print("No SD card");
  tft.display();
}

void show_wav_sel()
{
  int i;
  int j;
  char str[256];
  
  tft.fillRect(0, 0, _width, _height-9, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_GREEN);
  tft.setCursor(0, 0);
  tft.print("WAV(PCM 16bit) file");

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE);
  if (fnum<=0){
    tft.setCursor(0, 16);
    tft.print("No data");
    return;
  }

  tft.fillRect(0, 16+(num-disp_num)*8, _width, 8, ILI9163_RED);
  for(i=0;i+disp_num<fnum && i<MAX_LINE;i++){
    j=i+disp_num;
    tft.setCursor(0, 16+i*8);
    strcpy(str, fname[j]);
    str[21]='\0';
    tft.print(str);
  }
  tft.display();
}

void I2Splay(char *fname)
{
  int i;
  int val;
  int vol_int;
  int16_t *tmp;
  char str[32];
  size_t writesize = 0;

  tft.fillRect(0, 0, _width, _height-9, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE, ILI9163_BLACK);
  tft.setCursor(0, _height/2-16);
  tft.print(fname);
  tft.print("                                ");
  tft.display();

  file = SD.open(fname);  // 44100Hz, 16bit, stereo, linear PCM
  file.seek(offset);
  r_count=0;
  w_count=0;
  w_next=1;
  file_read=1;

  delay(500);

  while (1){
    if (r_count!=w_count || 
        (file_read==0 && r_count==w_count)){
      tmp=(int16_t *)data[r_count];
      vol_int=vol256[vol];
      if (I2S_stop>=1)
        vol_int=0;
//      for(i=0; i<data_num[r_count]; i++){
      for(i=0; i<BLOCK_SAMPLES*2; i++){
        val=(int)*(tmp++);
        val=val*vol_int/128;
        txbuf[i]=val;
      }
      pinMode(AMP_PIN, INPUT);
      writesize=BLOCK_SAMPLES*2*4;
      i2s_write( I2S_NUM_0, txbuf, BLOCK_SAMPLES*2*4, &writesize, portMAX_DELAY);              
      r_count++;
      if (r_count>=buf_max)
        r_count=0;
    }
    
    if (file_read==0 && r_count==w_count)
      break;
    if (I2S_stop>=4)
      break;

    if (I2S_stop>=1)
      I2S_stop++;
  }
  file_read=0;
  file.close();
  digitalWrite(LED_PIN, 0);

  // Sound stop
  pinMode(AMP_PIN, OUTPUT);
  digitalWrite(AMP_PIN, 0);
}
